/************************************************************************
 * uchroot.c                                                            
 *  by Jason F. McBrayer (after chroot.c by Roland McGrath)
 *  Copyright (C) 95, 96, 1997, 1999 Free Software Foundation, Inc.
 *
 *  Run a program, with the root directory set to a directory given on
 *  the command line.  Unlike standard chroot, uchroot is designed to
 *  be used by non-root users.  It is installed setuid root, and will
 *  drop privileges to the user's real uid after chrooting.  This
 *  _does_ have security implications.  I don't think that _in_effect_
 *  this program gives users any additional privileges except being
 *  able to chroot, but I admit to being a) not a security expert, and
 *  b) probably insufficiently paranoid.  This program should _not_
 *  under any circumstances be available from _inside_ a chroot jail.
 *
 *  I've tried very hard to avoid unexpected security problems like buffer
 *  overruns, but again, I'm not a security expert.
 *
 *  Only tested on Linux.  Likely to work on other GNU systems, may work
 *  on other POSIX systems.  If you want portability, you should probably
 *  hack on the chroot in the GNU shell utilities.  This code is mostly
 *  cribbed from there, anyway, therefore:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>

int usage();

int main(int argc, char **argv)
{
  uid_t uid, euid;
  char *shell, *program_name;

  program_name = argv[0];

  if (argc == 1)
    {
      fprintf(stderr,
              "Usage: %s <directory> [program]\n",
              program_name);
      exit(-1);
    }


  /* See if we would normally be able to access the directory. */
  if (access(argv[1], R_OK | X_OK))
    {
      fprintf(stderr,
	      "%s: can't access %s; error was \"%s\"\n",
	      program_name,
	      argv[1],
	      strerror(errno));
      exit(errno);
    }
  
  /* Try to do the chroot. */
  if (chroot(argv[1]))
    {
      fprintf(stderr,
	      "%s: couldn't change root directory to %s; error was \"%s\"\n",
	      program_name,
	      argv[1],
	      strerror(errno));
      exit(errno);
    }

  /* Try to do the chdir. */
  if (chdir("/"))
    {
      fprintf(stderr,
	      "%s: couldn't chdir to root directory; error was \"%s\"\n",
	      program_name,
	      strerror(errno));
      exit(errno);
    }
  
  /* Try to drop privileges */
  uid = getuid();
  euid = geteuid();
  if (euid != uid)
    {
      if (setuid(uid))
	{
	  fprintf(stderr,
		  "%s: couldn't drop privileges; error was \"%s\"\n",
		  program_name,
		  strerror(errno));
	  exit(errno);
	}
    }
#ifdef DEBUG
  else
    {
      fprintf(stderr,
	      "Warning: to get here we must be being run by root; not dropping\
 privileges.\n");
    }
#endif

  /* Try to exec the program, or an interactive shell. */
  if (argc == 2)
    {
      /* No command given; run an interactive shell */
      shell = getenv("SHELL");
      if (shell == NULL)
	shell = "/bin/sh";
      argv[0] = shell;
      argv[1] = "-i";
    }
  else
    {
      /* The command is in the following arguments */
      argv += 2;
    }

  execvp(argv[0], argv);
  fprintf(stderr,
	  "%s: cannot execute %s; error was \"%s\".\n",
	  program_name,
	  argv[0],
	  strerror(errno));
  exit(errno);
  return(errno);
}
